 /* Standard headers */
#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <stdbool.h>
#include <time.h>
#include <sys/times.h>
#include <syslog.h>
#include <signal.h>

/* Input subsystem interface */
#include <linux/input.h>
#include <linux/uinput.h>
#include <sys/ioctl.h>

/* Unix socket support */
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

/* KeysToCodes */
#include "keys2codes.h"


#define MYPORT "4950"    // the port users will be connecting to
#define MAXBUFLEN 2

int uinp_fd, sockfd;

// get sockaddr, IPv4 or IPv6:
void *get_in_addr( struct sockaddr *sa ) {
	if( sa -> sa_family == AF_INET ) {
		return &( ( ( struct sockaddr_in* ) sa ) -> sin_addr );
	}

	return &( ( ( struct sockaddr_in6* ) sa ) -> sin6_addr );
}

/* Setup the uinput device */
int setup_uinput_device() {
	char *uinp_device[] = { "/dev/uinput", "/dev/input/uinput", "/dev/misc/uinput" };
	int i, uinp_fd;
	for( i = 0; i < 3; i++ ) {
		// Try to open the uinput device
		if( ( uinp_fd = open( uinp_device[ i ], O_WRONLY | O_NDELAY ) ) <= 0 ) {
				continue;
		} else break;
	}
	if( uinp_fd <= 0 ) {
		printf( "Unable to open uinput-Device\n" );
		return -1;
	}
	printf( "Using %s\n", uinp_device[ i ] );

	struct uinput_user_dev uinp;
	memset( &uinp, 0, sizeof( uinp ) ); // Intialize the uInput device to NULL
	strncpy( uinp.name, "Virtual Keyboard Device", strlen(
		"Virtual Keyboard Device" ) );
	uinp.id.version = 4;
	uinp.id.bustype = BUS_USB;
	uinp.id.product = 1;
	uinp.id.vendor = 1;

	// Setup the uinput device
	ioctl( uinp_fd, UI_SET_EVBIT, EV_KEY );
	ioctl( uinp_fd, UI_SET_EVBIT, EV_REL );
	for( i = 0; i < 256; i++ ) {
		ioctl( uinp_fd, UI_SET_KEYBIT, i );
	}

	/* Create input device into input sub-system */
	write( uinp_fd, &uinp, sizeof( uinp ) );
	if( ioctl( uinp_fd, UI_DEV_CREATE ) ) {
		printf( "Unable to create UINPUT device." );
		return -1;
	}

	return uinp_fd;
}

void send_key_event( int fd, unsigned int keycode, int keyvalue ) {
	struct input_event event;
	gettimeofday( &event.time, NULL );

	event.type = EV_KEY;
	event.code = keycode;
	event.value = keyvalue;

	if( write( fd, &event, sizeof( event ) ) < 0 ) {
		printf( "simulate key error\n" );
	} else {
		printf( "simuate key %d, %d\n", keycode, keyvalue );
	}

	event.type = EV_SYN;
	event.code = SYN_REPORT;
	event.value = 0;
	write( fd, &event, sizeof( event ) );
	if( write( fd, &event, sizeof( event ) ) < 0 ) {
		printf( "simulate key error\n" );
	} else {
		printf( "simuate key %d, %d\n", keycode, keyvalue );
	}
}

int DestroyUinputDevice( int fd ) {
	if( ioctl( fd, UI_DEV_DESTROY ) == -1 ) {
		perror( "ioctl UI_DEV_DESTROY" );
		return -1;
	}
	return close( fd );
}

void DoSigTerm (int sig)
{
	close( sockfd );
	DestroyUinputDevice( uinp_fd );
	signal(sig,SIG_DFL);
	raise(sig);
}

int Key2Keycode( int key ) {
	if( KeysToCodes[ key ] != 0 ) {
		return( KeysToCodes[ key ] );
	} else {
		return( -1 );
	}
}

int main( int argc, char *argv[] ) {
	struct addrinfo hints, *servinfo, *p;
	int rv;
	int numbytes;
	struct sockaddr_storage their_addr;
	char buf[MAXBUFLEN];
	socklen_t addr_len;
	char s[INET6_ADDRSTRLEN];


// create and condition the universal input device.
	int uinp_fd;
	if( ( uinp_fd = setup_uinput_device() ) < 0 ) {
		printf( "Unable to find uinput device\n" );
		return -1;
	}

// Network-Stuff
	memset( &hints, 0, sizeof hints );
	hints.ai_family = AF_UNSPEC; // set to AF_INET to force IPv4
	hints.ai_socktype = SOCK_DGRAM;
	hints.ai_flags = AI_PASSIVE; // use my IP

	if( ( rv = getaddrinfo( NULL, MYPORT, &hints, &servinfo ) ) != 0 ) {
		fprintf( stderr, "getaddrinfo: %s\n", gai_strerror( rv ) );
		return 1;
	}

	// loop through all the results and bind to the first we can
	for( p = servinfo; p != NULL; p = p -> ai_next ) {
		if( ( sockfd = socket( p -> ai_family, p -> ai_socktype, p -> ai_protocol ) ) == -1) {
			perror( "listener: socket" );
			continue;
		}

		if( bind( sockfd, p -> ai_addr, p -> ai_addrlen ) == -1 ) {
			close( sockfd );
			perror( "listener: bind" );
			continue;
		}

		break;
	}

	if( p == NULL ) {
		fprintf( stderr, "listener: failed to bind socket\n" );
		return 2;
	}

	freeaddrinfo( servinfo );

	printf( "listener: waiting to recvfrom...\n" );

	addr_len = sizeof their_addr;

	signal( SIGTERM, DoSigTerm );

	do {
		if( ( numbytes = recvfrom( sockfd, buf, MAXBUFLEN - 1 , 0,
			( struct sockaddr * ) &their_addr, &addr_len ) ) == -1 ) {
			perror( "recvfrom" );
			exit( 1 );
		}

		printf( "listener: got packet from %s\n",
			inet_ntop( their_addr.ss_family,
				get_in_addr( ( struct sockaddr * ) &their_addr ),
				s, sizeof s ) );
		printf( "listener: packet is %d bytes long\n", numbytes );
		buf[numbytes] = '\0';
		printf( "listener: packet contains \"%s\"\n", buf );

		if( Key2Keycode( *buf ) > 0 ) {
			send_key_event( uinp_fd, Key2Keycode( *buf ), 1 );
			send_key_event( uinp_fd, Key2Keycode( *buf ), 0 );
		}

	} while( true );

	close( sockfd );
	DestroyUinputDevice( uinp_fd );
	return 0;
}
